/* File: calc.cpp
   Author: Richard Eisenberg (eir@cis)
   Desc: describes behavior of the calc window
*/

#include <QtGui>
#include "calc.h"
#include "input_validator.h"

// These are the definitions for the constants declared in calc.h
// Note that we don't use "static" here.
const QString Calc::PLUS = "+", Calc::MINUS = "-",
  Calc::TIMES = QChar(L'×'), Calc::DIVIDE = QChar(L'÷');
// A normal character in C++ is just 1 byte. This is not long enough
// to store Unicode characters, which include characters other than
// those on a standard English-language keyboard. So, to store characters
// like × or ÷ (the first one is not the letter x), we must prefix the
// character with L to say that it is a "long" or wide character.
// The explicit QChar constructor is necessary because there is no
// direct conversion from a wide character (wchar_t) to a QString;
// we have to go through QChar.

Calc::Calc()
{
  setWindowTitle("Qt Calc");

  // A layout tells how the individual widgets are laid out within the
  // window. The overall structure of the calculator is vertical, so
  // we use QVBoxLayout.
  QVBoxLayout* layout = new QVBoxLayout(this);
  
  // create the calculator display
  display = new QLineEdit(this);
  display->setAlignment(Qt::AlignRight);
  display->setReadOnly(true);
  
  // we use a child class of QValidator to make sure that the value in
  // the display is sensible (e.g. doesn't have two decimal points)
  display->setValidator(new InputValidator(this)); 
  layout->addWidget(display);

  // The lower part of the calculator is a grid, so we add a QGridLayout
  // to the QVBoxLayout
  QGridLayout* grid = new QGridLayout(this);
  layout->addLayout(grid);
  
  // By default, a clicked signal carries no extra information with it.
  // We wish to somehow package a number with the signals generated by
  // the number keys. This is exactly what a QSignalMapper does. When its
  // map() slot is triggered by some other object's signal, the QSignalMapper
  // signals "mapped(...)", where the ... is filled in with the value passed
  // into setMapping. See below.
  QSignalMapper* mapper = new QSignalMapper(this);

  for(int num = 1; num <= 9; num++)
  {
    QPushButton* btn = new QPushButton(QString::number(num), this);

    // instead of mapping the button directly to a slot in Calc, we connect
    // it to the mapper.
    connect(btn, SIGNAL(clicked()), mapper, SLOT(map()));
    
    // with this call, whenever the mapper gets signalled by btn, the mapper
    // will signal mapped(num). Because num is the value of the number of interest
    // we have successfully associated a number with each button's signal.
    mapper->setMapping(btn, num);

    // the math is just to get the positioning correct
    grid->addWidget(btn, 2 - (num - 1) / 3, (num - 1) % 3);
  }

  // now, we finally "close the loop" with the mapper -- we want the mapped(int)
  // signal to connect to number(int), saying that a number key has been pressed
  connect(mapper, SIGNAL(mapped(int)), this, SLOT(number(int)));

  QPushButton* zero = new QPushButton("0", this);
  connect(zero, SIGNAL(clicked()), mapper, SLOT(map()));
  mapper->setMapping(zero, 0);
  grid->addWidget(zero, 3, 0, 1, 2);

  QPushButton* dot = new QPushButton(".", this);
  connect(dot, SIGNAL(clicked()), this, SLOT(dot()));
  grid->addWidget(dot, 3, 2);

  // we do the same trick with the operator buttons, but here we pass
  // the operator strings instead of ints
  QSignalMapper* opMapper = new QSignalMapper(this);

  QString ops[] = { PLUS, MINUS, TIMES, DIVIDE };
  for(int i = 0; i < NUM_OPS; i++)
  {
    QPushButton* op = new QPushButton(ops[i], this);
    connect(op, SIGNAL(clicked()), opMapper, SLOT(map()));
    opMapper->setMapping(op, ops[i]);
    grid->addWidget(op, i, 3);
  }
  connect(opMapper, SIGNAL(mapped(QString)), this, SLOT(opButton(QString)));

  QPushButton* eq = new QPushButton("=", this);
  connect(eq, SIGNAL(clicked()), this, SLOT(equal()));
  grid->addWidget(eq, 4, 0, 1, 2);

  QPushButton* back = new QPushButton(QChar(L'←'), this);
  connect(back, SIGNAL(clicked()), this, SLOT(backspace()));
  grid->addWidget(back, 4, 2);

  QPushButton* clear = new QPushButton("C", this);
  connect(clear, SIGNAL(clicked()), this, SLOT(clear()));
  grid->addWidget(clear, 4, 3);

  display->setText("0");

  newNumber = true;
  lastNumber = 0;
  lastOp = "";
}

void Calc::maybeAppend(const QString& str)
{
  // newNumber is true when the next keypress erases what's in the display
  if(newNumber)
  {
    display->setText("0");
    newNumber = false;
  }

  // before updating the display, run it through the validator
  QString newText = display->text() + str;
  const QValidator* v = display->validator();
  int pos = 0; // needed for the validate call
  if(v->validate(newText, pos) == QValidator::Acceptable)
  {
    v->fixup(newText); // removes leading 0s
    display->setText(newText);
  }
}

void Calc::number(int n)
{
  maybeAppend(QString::number(n));
}

void Calc::dot()
{
  maybeAppend(".");
}

void Calc::performLastOp()
{
  double curValue = display->text().toDouble();
  
  // if there is a lastOp, perform the operation:
  if(lastOp == PLUS)
  {
    lastNumber += curValue;
  }
  else if(lastOp == MINUS)
  {
    lastNumber -= curValue;
  }
  else if(lastOp == TIMES)
  {
    lastNumber *= curValue;
  }
  else if(lastOp == DIVIDE)
  {
    lastNumber /= curValue;
  }
  else // there was no lastOp; just overwrite lastNumber
  {
    lastNumber = curValue;
  }
  
  display->setText(QString::number(lastNumber));
}

void Calc::opButton(QString op)
{
  // we've pressed an operator. Are we waiting for a new number? If so,
  // just update the lastOp:
  if(newNumber)
  {
    lastOp = op;
  }
  else
  {
    performLastOp();

    // record the operation, and we expect a new number
    lastOp = op;
    newNumber = true;
  }
}

void Calc::equal()
{
  // perform the operation and expect a new number:
  performLastOp();
  lastOp = "";
  newNumber = true;
}

void Calc::backspace()
{
  if(!newNumber) // only backspace while we're inputting
  {
    QString text = display->text();
    text.truncate(text.length() - 1);
    if(text.isEmpty())
    {
      text = "0";
    }

    display->setText(text);
  }
}

void Calc::clear()
{
  display->setText("0");
  lastNumber = 0;
  lastOp = "";
  newNumber = true;
}

// get the program off the ground
int main(int argc, char** argv)
{
  QApplication app(argc, argv);
  Calc c;
  c.show();
  return app.exec();
}
